using System;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.ComponentModel;

namespace Terragen2Sandbox
{
    public partial class Main : Form
    {
        Stream stream;
        BinaryReader bReader;
        StreamWriter pgmWriter;
        BinaryWriter rawWriter;

        string filePath;

        Bitmap bitmap; //Bitmap of elevation data
        Bitmap prevBitmap; //Scaled down preview image
        BitmapData bData;
        byte[] bDataArray;

        Progress progressDialog = null;

        ushort[,] aElevations; //Actual elevation data is stored here
        byte[,] Grays; //Preview image colors here

        ushort size; //Size of map

        int max = 0; //We need to know the maxinum value of heightmap for pgm writing

        bool operationCancel = false;

        //These are used to tell backworker what we want do
        enum taskMode { readTer, writePgm, readRaw, writeRaw };

        //----------------------------Terragen terrain stuff---------------------------------
        const string head = "TERRAGEN";
        const string head2 = "TERRAIN ";

        const string cSIZE = "SIZE";
        const string cSCAL = "SCAL";
        const string cXPTS = "XPTS";
        const string cYPTS = "YPTS";
        const string cCRAD = "CRAD";
        const string cCRVM = "CRVM";
        const string cALTW = "ALTW";

        long chunkBegin = 0;
        long prevPos;

        int xpts, ypts; //x-points, y-points
        float scalX, scalY, scalZ; //terrain scale (scalX = scalY = scalZ)
        float crad; //Thrash
        uint crvm; //Only used to notify if user is using sphererical map

        short HeightScale; //Thrash
        short BaseHeight; //Thrash
        int Elevation; //Elevation (heightmap value)
        //-----------------------------------------------------------------------------------

        //----------------------------Raw stuff---------------------------------------------
        long fileSize;
        long[] imgSizesInBytes = { 131072, 524288, 2097152, 8388608, 33554432, 134217728 }; //e.g. 512*512*2=524288
        ushort[] imgSizes = { 256, 512, 1024, 2048, 4096, 8192 };
        //-----------------------------------------------------------------------------------

        public Main()
        {
            InitializeComponent();
        }

        void ReadTerFile()
        {
            bReader = new BinaryReader(stream);

            if (ReadTerHead() != true) return;
            if (ReadTerChunks() != true) return;

            RunLongTask(taskMode.readTer);

            if (operationCancel == true)
            {
                aElevations = null;
                Grays = null;

                operationCancel = false;

                bReader.Close();
                stream.Close();

                return;
            }

            DrawPreview();

            bReader.Close();
            stream.Close();

            this.button2.Enabled = true;
            this.button4.Enabled = true;
        }

        bool ReadTerHead()
        {
            int i, ch;

            for (i = 0; i <= 7; i++)
            {
                ch = bReader.ReadChar();
                if(head[i] != ch)
                {
                    MessageBox.Show("Error: Not valid terragen file");
                    return false;
                }
            }
            for (i = 0; i <= 7; i++)
            {
                ch = bReader.ReadChar();
                if (head2[i] != ch)
                {
                    MessageBox.Show("Error: Not valid terrain file");
                    return false;
                }
            }
            chunkBegin = stream.Position;
            return true;
        }

        bool ReadTerChunks()
        {
            long pos;

            string fileInfo = "";

            prevPos = 0;

            //Read SIZE chunk
            pos = FindTerChunk(cSIZE);
            if (pos == -1)
            {
                    MessageBox.Show("Error reading data: Chunk SIZE not found");
                    return false;
            }
            stream.Seek(pos, SeekOrigin.Begin);
            size = bReader.ReadUInt16();

            prevPos = stream.Position;

            //Skip padding
            //stream.Seek(stream.Position + 2, SeekOrigin.Begin);

            //Read XPTS if exists
            pos = FindTerChunk(cXPTS);
            if (pos != -1)
            {
                stream.Seek(pos, SeekOrigin.Begin);
                xpts = bReader.ReadInt16();

                prevPos = stream.Position;
            }

            //Read YPTS if exists
            pos = FindTerChunk(cYPTS);
            if (pos != -1)
            {
                stream.Seek(pos, SeekOrigin.Begin);
                ypts = bReader.ReadInt16();

                prevPos = stream.Position;
            }

            if (xpts != ypts)
            {
                MessageBox.Show("Warning: Terrain is not square");
            }

            //Read SCAL if exists
            pos = FindTerChunk(cSCAL);
            if (pos != -1)
            {
                stream.Seek(pos, SeekOrigin.Begin);
                scalX = bReader.ReadSingle();
                scalY = bReader.ReadSingle();
                scalZ = bReader.ReadSingle();

                prevPos = stream.Position;

                if (scalX != scalY || scalY != scalZ || scalX != scalZ)
                {
                    MessageBox.Show("Error reading data: Terrain have non-uniform scaling");
                    return false;
                }
            }

            //Read CRAD if exists
            pos = FindTerChunk(cCRAD);
            if (pos != -1)
            {
                stream.Seek(pos, SeekOrigin.Begin);
                crad = bReader.ReadSingle();

                prevPos = stream.Position;
            }

            //Read CRVM if exists
            pos = FindTerChunk(cCRVM);
            if (pos != -1)
            {
                stream.Seek(pos, SeekOrigin.Begin);
                crvm = bReader.ReadUInt32();

                if (crvm != 0) MessageBox.Show("Terrain projected over sphere not supported, distortion may show up.");

                prevPos = stream.Position;
            }

            //Read ALTW chunk
            pos = FindTerChunk(cALTW);
            if (pos == -1)
            {
                MessageBox.Show("Error reading data: Chunk ALTW not found");
                return false;
            }
            stream.Seek(pos, SeekOrigin.Begin);

            HeightScale = bReader.ReadInt16();
            BaseHeight = bReader.ReadInt16();

            //Display terrain info
            fileInfo += "Map size: " + size + ", " + size + Environment.NewLine + Environment.NewLine;

            if (xpts == 0) fileInfo += "Num of x points: Not found" + Environment.NewLine + Environment.NewLine;
            else fileInfo += "Num of x points: " + xpts + Environment.NewLine + Environment.NewLine;
            
            if (ypts == 0) fileInfo += "Num of y points: Not found" + Environment.NewLine + Environment.NewLine;
            else fileInfo += "Num of y points: " + ypts + Environment.NewLine + Environment.NewLine;

            if (scalX == 0) fileInfo += "Metres per terrain unit: Not found" + Environment.NewLine + Environment.NewLine;
            else fileInfo += "Metres per terrain unit: " + (int)scalX + Environment.NewLine + Environment.NewLine;

            if (crad == 0) fileInfo += "Radius of planet: Not found" + Environment.NewLine + Environment.NewLine;
            else fileInfo += "Radius of planet: " + crad + Environment.NewLine + Environment.NewLine;

            fileInfo += "Spherical projection: " + (crvm == 0 ? "false" : "true") + Environment.NewLine + Environment.NewLine;

            fileInfo += "Height scale: " + HeightScale + Environment.NewLine + Environment.NewLine;

            fileInfo += "BaseHeight: " + BaseHeight + Environment.NewLine + Environment.NewLine;

            fileInfo += filePath;

            this.label1.Text = fileInfo;
            fileInfo = "";

            return true;
        }

        long FindTerChunk(string chunk)
        {
            int pos;
            string str = "";

            //HAX: Assuming that chunks dont't take over 100 bytes
            for (pos = 0; pos < 100; pos++)
            {
                stream.Seek(prevPos + pos, SeekOrigin.Begin);
                try
                {
                    str = new string(bReader.ReadChars(4));
                }
                catch
                {
                    str = "";
                    continue;
                }
                if (chunk == str)
                {
                    return stream.Position;
                }
            }
            return -1;
        }

        void ReadRawFile()
        {
            short i;
            bool sizeFound = false;
            bReader = new BinaryReader(stream);

            fileSize = stream.Length;

            for (i = 0; i < imgSizesInBytes.Length; i++)
            {
                if (fileSize == imgSizesInBytes[i])
                {
                    size = imgSizes[i];
                    sizeFound = true;
                }
            }
            if (sizeFound != true)
            {
                MessageBox.Show("Image must be square and power of two e.g. 1024x1024." + Environment.NewLine
                                + "Only 16-bit grayscale heightmaps supported.");
                return;
            }

            this.label1.Text = "Map size: " + size + ", " + size;

            RunLongTask(taskMode.readRaw);

            if (operationCancel == true)
            {
                aElevations = null;
                Grays = null;

                bReader.Close();
                stream.Close();

                operationCancel = false;

                return;
            }

            DrawPreview();

            this.button2.Enabled = true;
            this.button4.Enabled = true;

            bReader.Close();
            stream.Close();
        }

        void WritePgmFile()
        {
            pgmWriter = new StreamWriter(stream);

            pgmWriter.WriteLine("P2");
            pgmWriter.WriteLine(Convert.ToString(size) + " " + Convert.ToString(size));
            pgmWriter.WriteLine(Convert.ToString(max));

            RunLongTask(taskMode.writePgm);

            if (operationCancel == true)
            {
                pgmWriter.Close();
                stream.Close();

                File.Delete(filePath);

                operationCancel = false;
                return;
            }

            pgmWriter.Close();
            stream.Close();
        }

        void WriteRawFile()
        {
            rawWriter = new BinaryWriter(stream);

            RunLongTask(taskMode.writeRaw);

            if (operationCancel == true)
            {
                rawWriter.Close();
                stream.Close();

                File.Delete(filePath);

                operationCancel = false;
                return;
            }

            rawWriter.Close();
            stream.Close();
        }

        void RunLongTask(taskMode task)
        {
            progressDialog = new Progress();

            asyncWorker.RunWorkerAsync(task);

            progressDialog.ShowDialog(this);
            progressDialog = null;
        }

        //All the file reading and writing happens here on backworker because of long runtime and we want to
        //keep the ui responsive.
        void asyncWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            int updateInterval, i = 0;
            int x, y;

            //updateInterval = size/prograssbar maxinum
            updateInterval = (int)(size / 50);

            #region Read terrain heightdata
            if (e.Argument.ToString() == taskMode.readTer.ToString())
            {
                float k;

                //Ter file contains size-1 data points
                aElevations = new ushort[size + 1, size + 1];
                Grays = new byte[size + 1, size + 1];

                //normalization factor
                k = (float)(1.0 / 256);

                for (y = size; y > 0; y--)
                {
                    for (x = 0; x < size + 1; x++)
                    {
                        Elevation = bReader.ReadInt16();
                        //Elevation = (short)(BaseHeight + Elevation * HeightScale / 65536);

                        if (Elevation == -32768) Elevation = 0;
                        else Elevation = Elevation + 32768;

                        if (max < Elevation)
                        {
                            max = Elevation;
                        }

                        aElevations[y, x] = (ushort)Elevation; //Height data
                        Grays[y, x] = (byte)(Elevation * k); //Preview image

                        //Check if user cancelled
                        if (progressDialog.Cancel)
                        {
                            operationCancel = true;
                            e.Cancel = true;
                            return;
                        }
                    }

                    //Update the progress bar in another thread after we reached update interval (updates 50 times)
                    i++;
                    if (i >= updateInterval)
                    {
                        progressDialog.progressBar1.Invoke(
                            (MethodInvoker)delegate()
                            {
                                progressDialog.progressBar1.Value = Convert.ToInt32(Math.Abs(y - size) * (50.0 / (size)));
                            }
                        );
                        i = 0;
                    }
                }
            }
            #endregion

            #region Write pgm file
            if (e.Argument.ToString() == taskMode.writePgm.ToString())
            {
                short numValues = 0;

                //8 arvoa per rivi
                for (y = 0; y < size; y++)
                {
                    for (x = 0; x < size; x++)
                    {
                        pgmWriter.Write(Convert.ToString(aElevations[y, x]) + " ");

                        numValues++;
                        if (numValues > 7)
                        {
                            pgmWriter.Write(Environment.NewLine);
                            numValues = 0;
                        }
                        //Check if user cancelled
                        if (progressDialog.Cancel)
                        {
                            operationCancel = true;
                            e.Cancel = true;
                            return;
                        }
                    }
                    //Update the progress bar in another thread after we reached update interval (updates 50 times)
                    i++;
                    if (i >= updateInterval)
                    {
                        progressDialog.progressBar1.Invoke(
                            (MethodInvoker)delegate()
                            {
                                progressDialog.progressBar1.Value = Convert.ToInt32(y * (50.0 / (size)));
                            }
                        );
                        i = 0;
                    }
                }
            }
            #endregion

            #region Read raw data
            if (e.Argument.ToString() == taskMode.readRaw.ToString())
            {
                float k;

                aElevations = new ushort[size, size];
                Grays = new byte[size, size];

                //normalization factor
                k = (float)(1.0 / 256);

                for (y = 0; y < size; y++)
                {
                    for (x = 0; x < size; x++)
                    {
                        aElevations[y, x] = bReader.ReadUInt16();
                        //Elevation = (short)(BaseHeight + Elevation * HeightScale / 65536);

                        if (max < aElevations[y, x])
                        {
                            max = aElevations[y, x];
                        }

                        Grays[y, x] = (byte)(aElevations[y, x] * k); //Preview image

                        //Check if user cancelled
                        if (progressDialog.Cancel)
                        {
                            operationCancel = true;
                            e.Cancel = true;
                            return;
                        }
                    }

                    //Update the progress bar in another thread after we reached update interval (updates 50 times)
                    i++;
                    if (i >= updateInterval)
                    {
                        progressDialog.progressBar1.Invoke(
                            (MethodInvoker)delegate()
                            {
                                progressDialog.progressBar1.Value = Convert.ToInt32(y * (50.0 / (size)));
                            }
                        );
                        i = 0;
                    }
                }
            }
            #endregion

            #region Write raw data
            if (e.Argument.ToString() == taskMode.writeRaw.ToString())
            {
                for (y = 0; y < size; y++)
                {
                    for (x = 0; x < size; x++)
                    {
                        rawWriter.Write(aElevations[y, x]); //The most important thing here

                        //Check if user cancelled
                        if (progressDialog.Cancel)
                        {
                            operationCancel = true;
                            e.Cancel = true;
                            return;
                        }
                    }

                    //Update the progress bar in another thread after we reached update interval (updates 50 times)
                    i++;
                    if (i >= updateInterval)
                    {
                        progressDialog.progressBar1.Invoke(
                            (MethodInvoker)delegate()
                            {
                                progressDialog.progressBar1.Value = Convert.ToInt32(y * (50.0 / (size)));
                            }
                        );
                        i = 0;
                    }
                }
            }
            #endregion
        }

        void DrawPreview()
        {
            int bSize, x, y;
            int bOffset, pos = 0;

            bitmap = new Bitmap(size, size);
            bData = bitmap.LockBits(new Rectangle(0, 0, size, size), ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);

            bSize = bData.Height * bData.Stride;
            bOffset = bData.Stride - (bData.Width * 3);

            bDataArray = new byte[bSize];

            Marshal.Copy(bData.Scan0, bDataArray, 0, bSize);

            for (y = 0; y < size; y++)
            {
                for (x = 0; x < size; x++)
                {
                    pos = (y * bData.Stride) + (x * 3);
                    bDataArray[pos] = Grays[y, x];
                    pos += 1;
                    bDataArray[pos] = Grays[y, x];
                    pos += 1;
                    bDataArray[pos] = Grays[y, x];
                }
                pos += bOffset;
            }

            Marshal.Copy(bDataArray, 0, bData.Scan0, bSize);
            bDataArray = null;
            Grays = null;
            bitmap.UnlockBits(bData);

            prevBitmap = ResizeBitmap(bitmap, this.panel1.ClientSize.Width, this.panel1.ClientSize.Height);
            bitmap.Dispose();
            panel1.BackgroundImage = (Image)prevBitmap;
        }

        public Bitmap ResizeBitmap(Bitmap b, int nWidth, int nHeight)
        {
            Bitmap result = new Bitmap(nWidth, nHeight);
            using (Graphics g = Graphics.FromImage((Image)result))
                g.DrawImage(b, 0, 0, nWidth, nHeight);
            return result;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            this.openFileDialog1.Filter = ".ter file|*.ter";

            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                if ((stream = openFileDialog1.OpenFile()) != null)
                {
                    filePath = openFileDialog1.FileName;
                    ReadTerFile();
                }
                else MessageBox.Show("Error opening file");
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {
            this.saveFileDialog1.Filter = ".pgm file|*.pgm";

            if (this.saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                if ((stream = saveFileDialog1.OpenFile()) != null)
                {
                    filePath = saveFileDialog1.FileName;
                    WritePgmFile();
                }
                else MessageBox.Show("Error opening file");
            }
        }

        private void button3_Click(object sender, EventArgs e)
        {
            this.openFileDialog1.Filter = ".raw file|*.raw";

            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                if ((stream = openFileDialog1.OpenFile()) != null)
                {
                    filePath = openFileDialog1.FileName;
                    ReadRawFile();
                }
                else MessageBox.Show("Error opening file");
            }
        }

        private void button4_Click(object sender, EventArgs e)
        {
            this.saveFileDialog1.Filter = ".raw file|*.raw";

            if (this.saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                if ((stream = saveFileDialog1.OpenFile()) != null)
                {
                    filePath = saveFileDialog1.FileName;
                    WriteRawFile();
                }
                else MessageBox.Show("Error opening file");
            }
        }

        private void asyncWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if (progressDialog != null)
            {
                progressDialog.Hide();
                progressDialog = null;
            }

            if (e.Error != null)
            {
                MessageBox.Show(e.Error.Message);
                return;
            }
            if (e.Cancelled)
            {
                MessageBox.Show("Processing cancelled.");
                return;
            }
        }
    }
}